from time import time
import numpy as np
#import get_density
#import mgs_code
from . import get_density
from . import mgs_code

class mulguisin:
	def __init__(self, Rcut, x1, y1, z1=None, isort=None, boundaries=None, spherical=None, radius=None):
		self.Rcut = Rcut
		self.x1 = x1
		self.y1 = y1
		self.z1 = z1
		self.Nmgs = None
		self.imgs = None
		self.cng = None
		self.clm = None
		self.clg = None
		self.isort = isort
		self.boundaries = boundaries
		self.spherical = spherical
		self.radius = radius

	def get_isort(self):
		if self.spherical is None:
			if self.z1 is None:
				positions = np.vstack((self.x1,self.y1)).T
			else:
				positions = np.vstack((self.x1,self.y1,self.z1)).T
			print('Calculate Voronoi density')
			sta = time()
			if self.z1 is None:
				den = get_density.voronoi_2d_density(positions,self.boundaries)
				#den = voronoi_2d_density(positions,self.boundaries)
			else:
				den = get_density.voronoi_density(positions)
				#den = voronoi_density(positions)
			end = time()
			print('Calculation is done. Time = ', end - sta)
			isort = np.flip(den.argsort())
			return isort
		else:
			positions = np.vstack((self.x1,self.y1,self.z1)).T
			print('Calculate Spherical density')
			sta = time()
			den = get_density.spherical_density(positions,self.radius)
			end = time()
			print('Calculation is done. Time = ', end - sta)
			isort = np.flip(den.argsort())
			return isort

	def get_mgs(self):
		if self.isort is None:
			self.isort = self.get_isort()
		print('Calculate MGS')
		sta = time()
		if self.z1 is None:
			self.Nmgs,self.imgs,self.clg,self.clm,self.cng = mgs_code.mgs2d(self.Rcut, self.x1, self.y1, self.isort)
		else:
			self.Nmgs,self.imgs,self.clg,self.clm,self.cng = mgs_code.mgs3d(self.Rcut, self.x1, self.y1, self.z1, self.isort)
		end = time()
		print('Calculation is done. Time = ', end - sta)
		return self.Nmgs, self.imgs, self.clg, self.clm, self.cng

	def get_link(self):
		links = np.zeros((1,3))
		for i in range(self.Nmgs):
			for j in range(self.cng[i]):
				id_mom = self.clm[i][j]
				id_chi = self.clg[i][j]
				links = np.append(links,[[id_mom,id_chi,i]],axis=0)
		links = np.delete(links,0,0)
		return links

	def get_label(self):
		label = np.full((len(self.x1),), -1)
		for i in range(self.Nmgs):
			for j in range(self.cng[i]):
				label[self.clg[i][j]] = i
		return label
		

	def Get_TotLength(self,imgs=None):
		lsum = 0.
		for i in range(self.cng[imgs]):
			id_mom = self.clm[imgs][i]
			id_chi = self.clg[imgs][i]
			dist = (self.x1[id_mom]-self.x1[id_chi])**2 + (self.y1[id_mom]-self.y1[id_chi])**2 + (self.z1[id_mom]-self.z1[id_chi])**2
			lsum += np.sqrt(dist)
		return lsum

	def Get_Node(self,imgs=None,idgal=None):
		for i in range(self.cng[imgs]):
			if self.clg[imgs][i] == idgal:
				return self.clm[imgs][i]
	
	def Get_generation(self,imgs=None,idgal=None):
		ngen = 0
		id_chi = self.clg[imgs][idgal]
		id_mom = self.clg[imgs][0] # The origin of the generation, not mother galaxy
		while (id_chi != id_mom):
			id_node = self.Get_Node(imgs,id_chi)
			ngen+=1
			id_chi = id_node
		return ngen
	
	def Get_Length(self,imgs=None,idgal=None):
		lsum = 0.
		id_chi = self.clg[imgs][idgal]
		id_mom = self.clg[imgs][0]
		while (id_chi != id_mom):
			id_node = self.Get_Node(imgs,id_chi)
			dist = (self.x1[id_chi]-self.x1[id_node])**2 + (self.y1[id_chi]-self.y1[id_node])**2 + (self.z1[id_chi]-self.z1[id_node])**2
			lsum += np.sqrt(dist)
			id_chi = id_node
		return lsum
	
	def Get_Child(self,imgs=None,idgal=None):
		nchild = 0
		for i in range(idgal+1, self.cng[imgs]):
			if self.clm[imgs][i]==self.clg[imgs][idgal]: nchild+=1
		return nchild
	
	def Get_Degree(self,imgs=None,idgal=None):
		degree = 0
		id_mom = self.clg[imgs][0]
		for i in range(idgal+1,self.cng[imgs]):
			if self.clm[imgs][i] == self.clg[imgs][idgal]: degree+=1
		if self.clg[imgs][idgal] != id_mom: degree += 1
		return degree
	
	def Get_Branch(self,imgs=None):
		nbranch = 0
		for i in range(self.cng[imgs]):
			if self.Get_Child(imgs,i) == 0: nbranch += 1
		return nbranch
	
	def Get_OpenAngle(self,imgs=None,idgal=None):
		ang = -1
		id_chi = self.clg[imgs][idgal]
		id_mom = self.clg[imgs][0]
		if id_chi!=id_mom:
			id_node = self.Get_Node(imgs,id_chi)
			if id_node!=id_mom:
				id_axis = self.Get_Node(imgs,id_node)
				ang = self.Get_Angle(self.x1[id_chi],self.y1[id_chi],self.z1[id_chi],self.x1[id_node],self.y1[id_node],self.z1[id_node],self.x1[id_axis],self.y1[id_axis],self.z1[id_axis])
		return ang
	
	def Get_LinkPolarAngle(self,imgs=None,idgal=None):
		ang = -1
		id_chi = self.clg[imgs][idgal]
		id_mom = self.clg[imgs][0]
		if id_chi!=id_mom:
			id_node = self.Get_Node(imgs,id_chi)
			ang = self.Get_Angle(self.x1[id_chi],self.y1[id_chi],self.z1[id_chi],self.x1[id_node],self.y1[id_node],self.z1[id_node],0,0,0)
		return ang

	def Get_PolarAngle(self,imgs=None,idgal=None):
		ang = -1
		id_chi = self.clg[imgs][idgal]
		id_mom = self.clg[imgs][0]
		if id_chi!=id_mom:
			ang = self.Get_Angle(self.x1[id_chi],self.y1[id_chi],self.z1[id_chi],self.x1[id_mom],self.y1[id_mom],self.z1[id_mom],0,0,0)
		return ang
	
	def Get_Angle(self,x3=None,y3=None,z3=None,x2=None,y2=None,z2=None,x1=None,y1=None,z1=None):
		da = np.sqrt((x3-x2)**2 + (y3-y2)**2 + (z3-z2)**2)
		db = np.sqrt((x2-x1)**2 + (y2-y1)**2 + (z2-z1)**2)
		dc = (x3-x1)**2 + (y3-y1)**2 + (z3-z1)**2
		cosalpha = (da**2 + db**2 - dc)/(2.*da*db)
		return np.arccos(cosalpha)



from itertools import permutations

import numpy as np
from numpy.linalg import det
from scipy.spatial import Delaunay, Voronoi, distance, voronoi_plot_2d
from scipy.spatial import cKDTree


#----------------------------------
def tetravol(a,b,c,d):
    '''Calculates the volume of a tetrahedron, given vertices a,b,c,d'''
    tetravol=abs(np.dot((a-d),np.cross((b-d),(c-d))))/6
    return tetravol

def triangle_area(a,b,c):
    " Ref : https://kr.mathworks.com/matlabcentral/answers/119984-how-to-find-the-area-of-the-triangles-formed-as-a-result-of-delaunay-triangulation"
    #area = abs((x2-x1)*(y3-y1) - (x3-x1)*(y2-y1))
    area = abs((b[0]-a[0])*(c[1]-a[1]) - (c[0]-a[0])*(b[1]-a[1]))/2.
    return area

def intersection2(p1,p2,p3,p4):
    " Ref : https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection"
    t = det([[p1[0]-p3[0],p3[0]-p4[0]],[p1[1]-p3[1],p3[1]-p4[1]]])/det([[p1[0]-p2[0],p3[0]-p4[0]],[p1[1]-p2[1],p3[1]-p4[1]]])
    u = det([[p1[0]-p3[0],p1[0]-p2[0]],[p1[1]-p3[1],p1[1]-p2[1]]])/det([[p1[0]-p2[0],p3[0]-p4[0]],[p1[1]-p2[1],p3[1]-p4[1]]])

    return [[p1[0]+t*(p2[0]-p1[0]), p1[1]+t*(p2[1]-p1[1])],[p3[0]+u*(p4[0]-p3[0]),p3[1]+u*(p4[1]-p3[1])]]

def vol(vor,p):
    '''Calculate volume of 3d Voronoi cell based on point p.'''
    dpoints=[]
    vol=0
    for v in vor.regions[vor.point_region[p]]:
        dpoints.append(list(vor.vertices[v]))
    tri=Delaunay(np.array(dpoints),qhull_options='Q12 Qs Qz')
    for simplex in tri.simplices:
        vol+=tetravol(np.array(dpoints[simplex[0]]),np.array(dpoints[simplex[1]]),np.array(dpoints[simplex[2]]),np.array(dpoints[simplex[3]]))
    return vol

"""
Below codes are for 2D case. 
"""

def find_boundary(vertex,mid_point,x1,x2,y1,y2,angle):
    if angle > 45. and angle < 135.:
        intersect = intersection2(vertex,mid_point,[x1,y2],[x2,y2]) # upper boundary
    elif angle > 135. or angle < -135.:
        intersect = intersection2(vertex,mid_point,[x1,y1],[x1,y2]) # left boundary
    elif angle > -135. and angle < -45.:
        intersect = intersection2(vertex,mid_point,[x1,y1],[x2,y1]) # bottom boundary
    elif angle > -45. or angle < 45.:
        intersect = intersection2(vertex,mid_point,[x2,y1],[x2,y2]) # right boundary
    elif angle == 45.:
        intersect = np.array([x2,y2])
    elif angle == 135.:
        intersect = np.array([x1,y2])
    elif angle == -135.:
        intersect = np.array([x1,y1])
    elif angle == -45.:
        intersect = np.array([x2,y1])
    else:
        print('Calculation of angle is wrong!')
        intersect = [np.inf,np.inf]
    return intersect


"""
Now code can find far_point. I borrow the code that calculate the far_point from
"scipy : voronoi_plot_2d" function.
"""
def area(vor,p,boundaries):
    '''Calculate area of 2D Voronoi cell based on point p'''
    area = 0
    x1,x2,y1,y2 = boundaries[0],boundaries[1],boundaries[2],boundaries[3]
    ind_vert_in_cell = vor.regions[vor.point_region[p]].copy()
    #print('length of index vertices in a cell : ',len(ind_vert_in_cell))
    center = vor.points.mean(axis=0)
    add_vertices = []

    if  -1 in ind_vert_in_cell:
        test_ridge_vertices = vor.ridge_vertices.copy()
        test_ridge_point = vor.ridge_points.copy()
        for permute in list(permutations(ind_vert_in_cell,2)):
            if -1 in permute: # e.g. [-1, 25 ] or [196, -1]
                try:
                    #print(permute)
                    ind_vert = test_ridge_vertices.index(list(permute))
                    ind_point = test_ridge_point[ind_vert]
                    #print('ridge point : ',ind_point)
                    # Find tangent vector of two ridge points
                    mid1 = vor.points[ind_point[0]]
                    mid2 = vor.points[ind_point[1]]
                    t = mid2-mid1
                    t /= np.linalg.norm(t)
                    n = np.array([-t[1],t[0]]) # Normal vector

                    mid_point = [(mid1[0]+mid2[0])/2. , (mid1[1]+mid2[1])/2.]
                    #print('median point : ',mid_point)

                    # Find direction
                    direction = np.sign(np.dot(mid_point - center, n)) * n
                    # Find boundary
                    linked_index = list(permute)[1]
                    vertex = list(vor.vertices[linked_index])
                    #print('linked vertex : ',vertex)
                    far_point = vertex + direction
                    # check where is the base?
                    test_vector = np.array([1.,0.])
                    inner_product = np.dot(test_vector,direction)
                    if inner_product < 0:
                        angle = -np.arccos(inner_product)*180./np.pi
                    else:
                        angle = np.arccos(inner_product)*180./np.pi
                    # First, check the postion of vertex in quadrant

                    intersect = find_boundary(vertex,far_point,x1,x2,y1,y2,angle)

                    add_vertices.append(intersect[0])
                except:
                    pass
    else:
        pass

    #print(ind_vert_in_cell)

    if -1 in ind_vert_in_cell:
        ind_vert_in_cell.remove(-1)
    origin_vertice = vor.vertices[ind_vert_in_cell].tolist()
    new_vertices = origin_vertice.copy()
    new_vertices = new_vertices + add_vertices

    area = 0
    tri=Delaunay(np.array(new_vertices),qhull_options='Q12 Qs Qz')
    for simplex in tri.simplices:
        area+=triangle_area(new_vertices[simplex[0]],new_vertices[simplex[1]],new_vertices[simplex[2]])
    #print(area)
    return area
    
def voronoi_2d_density(positions,boundaries):
    galvor=Voronoi(positions)
    galden=np.zeros((len(positions),))
    for i in range(len(positions)):
        galden[i]=1./area(galvor,i,boundaries)
    return galden

def voronoi_density(positions):
    galvor=Voronoi(positions)
    galden=np.zeros((len(positions),))
    for i in range(len(positions)):
        galden[i]=1./vol(galvor,i)
    return galden

def spherical_density(positions,radius):
    galden=np.zeros((len(positions),))
    volume = 4.*np.pi*radius*radius*radius/3.
    tree = cKDTree(positions)
    for i in range(len(positions)):
       num_gal = tree.query_ball_point(positions[i], r=radius,return_length=True)
       galden[i] = num_gal/volume
    return galden
       
    



from setuptools import setup, find_packages

setup(
	name='Mulguisin',
	version='0.9',
	license='BSD',
	description='Mulguisin: Cluster/group finding algorithm',
	author='Young Ju, Inkyu Park, Sungwook E. Hong, Cristiano G. Sabiu',
	author_email='youngju@uos.ac.kr',
	packages=find_packages(),
	#package_dir={'': 'Mulguisin'},
	url='https://github.com/youngju20/Mulguisin',
	install_requires=[
		'scipy','numpy','h5py','numba',
	],
		
)



from numba import njit
from numba.typed import List

@njit
def mgs3d(Rcut,x1,y1,z1,isort):
	Nmgs=0
	imgs = List()
	cng = List()
	clg = List()
	clm = List()
	#cll = List()
	for i in range(len(x1)):
		ic = isort[i]
		Icc=-1
		Iwg=-1
		Dmin=999

		for ig in range(0,Nmgs):
			Dist=999
			for icl in range(0,cng[ig]):
				itmp=clg[ig][icl]
				d2 = (x1[ic]-x1[itmp])**2+(y1[ic]-y1[itmp])**2+(z1[ic]-z1[itmp])**2
				if (d2<Dist): 
					idmin=itmp
					Dist=d2
			if (Dist < Dmin):
				Iwg=ig
				Icc=idmin
				Dmin=Dist
		if (Dmin < Rcut*Rcut):
			imgs.append(Iwg)
			clg[Iwg].append(ic)
			clm[Iwg].append(Icc)
			cng[Iwg]+=1
		else:
			imgs.append(Nmgs)	
			clg.append([ic])
			clm.append([ic])
			cng.append(1)
			Nmgs+=1
	return Nmgs,imgs,clg,clm,cng

@njit
def mgs2d(Rcut,x1,y1,isort):
	Nmgs=0
	imgs = List()
	cng = List()
	clg = List()
	clm = List()
	#cll = List()
	for i in range(len(x1)):
		ic = isort[i]
		Icc=-1
		Iwg=-1
		Dmin=999

		for ig in range(0,Nmgs):
			Dist=999
			for icl in range(0,cng[ig]):
				itmp=clg[ig][icl]
				d2 = (x1[ic]-x1[itmp])**2+(y1[ic]-y1[itmp])**2
				if (d2<Dist): 
					idmin=itmp
					Dist=d2
			if (Dist < Dmin):
				Iwg=ig
				Icc=idmin
				Dmin=Dist
		if (Dmin < Rcut*Rcut):
			imgs.append(Iwg)
			clg[Iwg].append(ic)
			clm[Iwg].append(Icc)
			cng[Iwg]+=1
		else:
			imgs.append(Nmgs)	
			clg.append([ic])
			clm.append([ic])
			cng.append(1)
			Nmgs+=1
	return Nmgs,imgs,clg,clm,cng
